﻿using System.Collections.Generic;
using ColossalFramework.Globalization;
using ICities;
using UEObj = UnityEngine.Object;

namespace ${namespace}
{
	/// <summary>
	/// Customizes electricity building stats when a city is loaded.
	/// </summary>
	public class Loading${namespace} : LoadingExtensionBase
	{
		/// <summary>
		/// Runs when the addon is created.
		/// </summary>
		public override void OnCreated(ILoading loading)
		{
			base.OnCreated(loading);
			Log.Message("OnCreated");

			// When debugging, it is helpful to run CustomizeStats() every time the addon is recompiled and reloaded:
			//CustomizeStats();
		}

		/// <summary>
		/// Runs when a city is loaded.
		/// </summary>
		public override void OnLevelLoaded(LoadMode mode)
		{
			base.OnLevelLoaded(mode);
			Log.Message("OnLevelLoaded");
			CustomizeStats();
		}

		#region Helpers
		/// <summary>
		/// Customizes the stats for all power plants.
		/// </summary>
		private void CustomizeStats()
		{
			var buildingInfos = FindBuildingsOfClass(ItemClass.Service.Electricity);
			CustomizeBuildings(buildingInfos);

			var netInfos = FindNetsOfClass(ItemClass.Service.Electricity);
			CustomizeNets(netInfos);

			RefreshElectricityGroupPanel();
		}

		/// <summary>
		/// Refreshes the power plant chooser toolbar to reflect the customized stats.
		/// </summary>
		private static void RefreshElectricityGroupPanel()
		{
			var panel = UEObj.FindObjectOfType<ElectricityGroupPanel>();
			panel.RefreshPanel();
		}

		/// <summary>
		/// Finds all building prototypes matching the provided <paramref="service"/>.
		/// </summary>
		/// <param name="service">The type of building.</param>
		/// <returns>All matching building prototypes.</returns>
		public static IList<BuildingInfo> FindBuildingsOfClass(ItemClass.Service service)
		{
			var count = PrefabCollection<BuildingInfo>.LoadedCount();
			var list = new List<BuildingInfo>();
			for (uint i = 0; i < count; ++i)
			{
				var loaded = PrefabCollection<BuildingInfo>.GetLoaded(i);
				if (loaded.m_class.m_service == service)
				{
					list.Add(loaded);
				}
			}
			return list;
		}

		/// <summary>
		/// Finds all network prototypes matching the provided <paramref="service"/>.
		/// </summary>
		/// <remarks>
		/// Network buildings are things like roads, electrical lines, and dams.
		/// </remarks>
		/// <param name="service">The type of building.</param>
		/// <returns>All matching building prototypes.</returns>
		public static IList<NetInfo> FindNetsOfClass(ItemClass.Service service)
		{
			var count = PrefabCollection<NetInfo>.LoadedCount();
			var list = new List<NetInfo>();
			for (uint i = 0; i < count; ++i)
			{
				var loaded = PrefabCollection<NetInfo>.GetLoaded(i);
				if (loaded.m_class.m_service == service)
				{
					list.Add(loaded);
				}
			}
			return list;
		}

		/// <summary>
		/// Returns the localized name of a building.
		/// </summary>
		/// <remarks>
		/// This method is useful mainly because user-created buildings will have a long, mangled name.
		/// </remarks>
		/// <param name="info">The building prototype</param>
		/// <returns>The localized name of the building, or (if not found), the non-localized name.</returns>
		private static KeyValuePair<string, string> GetBuildingNames(BuildingInfo info)
		{
			if (info.m_prefabDataIndex == -1)
			{
				return new KeyValuePair<string, string>(info.name, info.name);
			}

			var preFabName = PrefabCollection<BuildingInfo>.PrefabName((uint) info.m_prefabDataIndex);

			//Log.Message(string.Format("GetBuildingName: Getting locale name for {0}, {1}, {2}", info.m_prefabDataIndex, info.name, preFabName));
			var localizedName = Locale.Exists("BUILDING_TITLE", preFabName)
				? Locale.Get("BUILDING_TITLE", preFabName)
				: preFabName;

			return new KeyValuePair<string, string>(preFabName, localizedName);
		}

		/// <summary>
		/// Returns the localized name of a network.
		/// </summary>
		/// <remarks>
		/// This method is useful mainly because user-created networks will have a long, mangled name.
		/// </remarks>
		/// <param name="info">The network prototype</param>
		/// <returns>The localized name of the network, or (if not found), the non-localized name.</returns>
		private static KeyValuePair<string, string> GetNetNames(NetInfo info)
		{
			if (info.m_prefabDataIndex == -1)
			{
				return new KeyValuePair<string, string>(info.name, info.name);
			}

			var preFabName = PrefabCollection<BuildingInfo>.PrefabName((uint) info.m_prefabDataIndex);

			//Log.Message(string.Format("GetNetName: Getting locale name for {0}, {1}, {2}", info.m_prefabDataIndex, info.name, preFabName));
			var localizedName = Locale.Exists("NET_TITLE", preFabName)
				? Locale.Get("NET_TITLE", preFabName)
				: preFabName;

			return new KeyValuePair<string, string>(preFabName, localizedName);
		}
		#endregion Helpers

		#region Customize
		/// <summary>
		/// Customizes the stats of all provided <paramref="buildingInfos"/>.
		/// </summary>
		/// <param name="buildingInfos">A list of building prototypes.</param>
		private void CustomizeBuildings(IList<BuildingInfo> buildingInfos)
		{
			foreach (var buildingInfo in buildingInfos)
			{
				var service = buildingInfo.m_class.m_service;
				switch (service)
				{
					case ItemClass.Service.Electricity:
						break;
					default:
						//Log.Message("BoP skipping service: " + service + ", name: " + buildingInfo.name);
						continue;
				}


				if (buildingInfo.m_buildingAI is PowerPlantAI)
				{
					var names = GetBuildingNames(buildingInfo);
					var plant = (PowerPlantAI) buildingInfo.m_buildingAI;
					Rebuilder.Modify(names.Key, names.Value, plant);
				}
			}
		}

		/// <summary>
		/// Customizes the stats of all provided <paramref="netInfos"/>.
		/// </summary>
		/// <param name="netInfos">A list of network prototypes.</param>
		private void CustomizeNets(IList<NetInfo> netInfos)
		{
			foreach (var netInfo in netInfos)
			{
				var service = netInfo.m_class.m_service;
				switch (service)
				{
					case ItemClass.Service.Electricity:
						break;
					default:
						//Log.Message("BoP skipping service: " + service + ", name: " + netInfo.name);
						continue;
				}

				if (netInfo.m_netAI is DamAI)
				{
					var names = GetNetNames(netInfo);
					var dam = (DamAI) netInfo.m_netAI;
					Rebuilder.Modify(names.Key, names.Value, dam);
				}
			}
		}
		#endregion Customize
	}
}