﻿using ColossalFramework;
using ColossalFramework.UI;
using ICities;
using UEObject = UnityEngine.Object;

namespace ${namespace}
{
	/// <summary>
	/// Displays electricity building resource durations for non-standard resource types (Grain, Lumber, Ore).
	/// </summary>
	public class Threading${namespace} : ThreadingExtensionBase
	{
		/// <summary>
		/// The name of the electricity facility info panel.
		/// </summary>
		private const string ElectricityInfoPanelName = "(Library) CityServiceWorldInfoPanel";

		/// <summary>
		/// The name of the info label on the electricity facility info panel.
		/// </summary>
		private const string ElectricityInfoPanelInfoLabelName = "Info";

		/// <summary>
		/// The name of the custom label that this extension will add to the electricity facility info panel.
		/// </summary>
		private const string ResourceSupplyLabelName = "${namespace}.AltResourceSupply";

		/// <summary>
		/// True if this extension has initialized, otherwise false.
		/// </summary>
		private bool _isInitialized;

		/// <summary>
		/// A reference to the electricity facility info panel, if it could be found.
		/// </summary>
		private UIPanel _infoPanel;

		/// <summary>
		/// A reference to the electricity facility info panel's info label, if it could be found.
		/// </summary>
		private UILabel _infoLabel;

		/// <summary>
		/// The custom label that this extension added to the electricity facility info panel, if it could be created.
		/// </summary>
		private UILabel _resourceSupplyLabel;

		/// <summary>
		/// Disposes of the custom label added to the electricity facility info panel.
		/// </summary>
		public override void OnReleased()
		{
			//Log.Message("Threading${namespace} OnReleased");
			if (_infoPanel == null)
				return;

			if (_resourceSupplyLabel != null)
			{
				// Dispose of the custom label.
				_infoPanel.RemoveUIComponent(_resourceSupplyLabel);
				UEObject.Destroy(_resourceSupplyLabel);
				_resourceSupplyLabel = null;
			}
			_infoPanel = null;
		}

		/// <summary>
		/// Initializes the view, if necessary, and then updates the view.
		/// </summary>
		/// <param name="realTimeDelta">The amount of real time that has passed since the last update</param>
		/// <param name="simulationTimeDelta">The amount of simulation time that has passed since the last update.</param>
		public override void OnUpdate(float realTimeDelta, float simulationTimeDelta)
		{
			base.OnUpdate(realTimeDelta, simulationTimeDelta);

			if (!_isInitialized)
			{
				InitializeView();
			}

			UpdateView();
		}

		/// <summary>
		/// Updates the view on the electricity facility info panel, if necessary.
		/// </summary>
		private void UpdateView()
		{
			if (_infoPanel == null || !_infoPanel.isVisible || _resourceSupplyLabel == null)
				return;

			_resourceSupplyLabel.isVisible = false;
			_resourceSupplyLabel.text = string.Empty;

			var buildingIndex = WorldInfoPanel.GetCurrentInstanceID().Building;

			var buildingManager = Singleton<BuildingManager>.instance;
			var building = buildingManager.m_buildings.m_buffer[buildingIndex];
			var buildingInfo = building.Info;

			var powerPlantAi = buildingInfo.m_buildingAI as PowerPlantAI;
			if (powerPlantAi == null)
				return;

			string resourceName;
			switch (powerPlantAi.m_resourceType)
			{
				case TransferManager.TransferReason.Ore:
					// The game treats Ore and Coal the same, for the most part.
					// However, it is hard-coded to only show resource duration for Coal.
					// Therefore, it is safe to pretend that Ore is actually Uranium.
					resourceName = "Uranium";
					break;
				case TransferManager.TransferReason.Grain:
					resourceName = "Grain";
					break;
				case TransferManager.TransferReason.Logs:
					// The game treats Logs and Lumber the same, for the most part.
					// Let's make a small distinction here: Logs = Wood, Lumber = Wood pellets.
					resourceName = "Wood";
					break;
				case TransferManager.TransferReason.Lumber:
					// The game treats Logs and Lumber the same, for the most part.
					// Let's make a small distinction here: Logs = Wood, Lumber = Wood pellets.
					resourceName = "Wood pellets";
					break;
				case TransferManager.TransferReason.Oil:
					// The game treats Oil and Petrol the same, for the most part.
					// However, it is hard-coded to only show resource duration for Petrol.
					// Therefore, it is safe to pretend that Oil is actually Natural gas.
					resourceName = "Natural gas";
					break;
				default:
					return;
			}

			var resourceDuration = powerPlantAi.GetResourceDuration(buildingIndex, ref building);

			const string localizedString = "{0} stored for {1} weeks.";
			var formattedString = string.Format(localizedString, resourceName, resourceDuration);

			_infoLabel.autoHeight = true;
			_resourceSupplyLabel.text = formattedString;
			_resourceSupplyLabel.isVisible = true;
		}

		/// <summary>
		/// Initializes the custom label on the electricity facility info panel.
		/// </summary>
		private void InitializeView()
		{
			_isInitialized = true;

			if (_infoPanel == null)
			{
				_infoPanel = UIView.Find<UIPanel>(ElectricityInfoPanelName);

				if (_infoPanel == null)
					return;

				//Log.Message("Found (Library) CityServiceWorldInfoPanel");
			}

			_infoLabel = _infoPanel.Find<UILabel>(ElectricityInfoPanelInfoLabelName);
			var existingLabel = _infoPanel.Find<UILabel>(ResourceSupplyLabelName);
			if (existingLabel != null)
			{
				_infoPanel.RemoveUIComponent(existingLabel);
				UEObject.Destroy(existingLabel);
			}

			_resourceSupplyLabel = _infoPanel.AddUIComponent<UILabel>();
			_resourceSupplyLabel.name = ResourceSupplyLabelName;

			_resourceSupplyLabel.AlignTo(_infoLabel, UIAlignAnchor.BottomLeft);

			_resourceSupplyLabel.font = _infoLabel.font;
			_resourceSupplyLabel.textScale = _infoLabel.textScale;

			_resourceSupplyLabel.color = _infoLabel.color;
			_resourceSupplyLabel.textColor = _infoLabel.textColor;
			_resourceSupplyLabel.dropShadowColor = _infoLabel.dropShadowColor;
			_resourceSupplyLabel.bottomColor = _infoLabel.bottomColor;
		}
	}
}