﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.IO;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Text.RegularExpressions;
using DocumentFormat.OpenXml.Packaging;
using DocumentFormat.OpenXml.Spreadsheet;

namespace PowerGenerator
{
	public static class Worker
	{
		/// <summary>
		/// Generates the addon files from the Excel file.
		/// </summary>
		/// <param name="excelFilePath">The path to the Excel file.</param>
		/// <param name="excelSheetName">The name of the sheet to read from in the Excel file.</param>
		/// <param name="includeDifferencesOnly">If true, will include only the fields that were changed from the defaults.</param>
		/// <param name="outputPath">The output directory for the addon.</param>
		/// <param name="addonName">The name of the addon.</param>
		/// <param name="description">A long description of the addon.</param>
		/// <param name="addonNamespace">The .NET namespace to use for the addon.</param>
		public static void Generate(
			string excelFilePath, string excelSheetName, bool includeDifferencesOnly,
			string outputPath, string addonName,
			string description, string addonNamespace)
		{
			if (string.IsNullOrWhiteSpace(excelFilePath))
				throw new ArgumentNullException("excelFilePath", "The Excel file is required");
			if (string.IsNullOrWhiteSpace(excelSheetName))
				throw new ArgumentNullException("excelSheetName", "The Excel sheet name is required");
			if (string.IsNullOrWhiteSpace(outputPath))
				throw new ArgumentNullException("excelFilePath", "The output path is required");
			if (string.IsNullOrWhiteSpace(addonName))
				throw new ArgumentNullException("addonName", "The addon name is required");
			if (string.IsNullOrWhiteSpace(description))
				throw new ArgumentNullException("description", "The addon description is required");
			if (string.IsNullOrWhiteSpace(addonNamespace))
				throw new ArgumentNullException("addonNamespace", "The addon namespace is required");

			if (!Regex.IsMatch(addonNamespace, @"\A\w+\Z"))
				throw new ArgumentException("Namespace must be letters only", "addonNamespace");

			DirectoryInfo outputDirectory;
			try
			{
				outputDirectory = Directory.CreateDirectory(outputPath);
			}
			catch (Exception ex)
			{
				throw new Exception("Unable to create the output directory " + outputPath, ex);
			}

			var templateParameters = new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase)
			{
				{ TemplateAddonNamespace, addonNamespace },
				{ TemplateAddonName, EscapeVerbatimString(addonName) },
				{ TemplateDescription, EscapeVerbatimString(description) }
			};

			var allStats = GetTrimmedTranslatedStats(excelFilePath, excelSheetName, includeDifferencesOnly, DefaultSheetName);

			var code = GenerateRebuilderCode(allStats);
			foreach (var x in code)
			{
				templateParameters.Add(x.Key, x.Value);
			}

			// Get all templates out of the assembly and run keyword replacement on them.
			var assembly = Assembly.GetExecutingAssembly();
			var resourceNames = assembly.GetManifestResourceNames();
			var namePattern = new Regex(@"\A(\w+\.)+Templates\.", RegexOptions.IgnoreCase);
			var nameLookup = resourceNames
				.Where(s => namePattern.IsMatch(s))
				.ToDictionary(s => namePattern.Replace(s, string.Empty), StringComparer.InvariantCultureIgnoreCase);

			var lineBreaks = new Regex(@"(\n\r?|\r\n?)", RegexOptions.CultureInvariant);
			foreach (var resourceName in nameLookup)
			{
				string fileContents;
				using (var stream = assembly.GetManifestResourceStream(resourceName.Value))
				{
					if (stream == null)
						throw new Exception("Unable to open the template stream for " + resourceName.Key);

					using (var reader = new StreamReader(stream))
					{
						fileContents = reader.ReadToEnd();
					}
				}

				fileContents = FormatString(fileContents, templateParameters);
				fileContents = lineBreaks.Replace(fileContents, Environment.NewLine);

				var outputFile = new FileInfo(Path.Combine(outputDirectory.FullName, resourceName.Key));
				using (var outputFileStream = outputFile.Open(FileMode.Create, FileAccess.Write, FileShare.None))
				using (var writer = new StreamWriter(outputFileStream))
				{
					writer.Write(fileContents);
				}
			}
		}

		/// <summary>
		/// Retrieves the trimmed, translated stats from an Excel file.
		/// Any stats that are the same between sheet <paramref name="statsSheetName"/> and the DEFAULTS sheet
		/// will be discarded, because they don't need to be changed.
		/// </summary>
		/// <param name="excelFilePath">The path to the Excel file.</param>
		/// <param name="statsSheetName">The name of the stats sheet to read from in the Excel file.</param>
		/// <param name="includeDifferencesOnly">If true, will include only the fields that were changed from the defaults.</param>
		/// <param name="defaultsSheetName">The name of the defaults sheet to read from in the Excel file.</param>
		/// <returns>A dictionary of trimmed, translated stats keyed on power plant names.</returns>
		private static Dictionary<string, Dictionary<string, string>> GetTrimmedTranslatedStats(string excelFilePath,
			string statsSheetName, bool includeDifferencesOnly, string defaultsSheetName)
		{
			FileInfo tempInputFile = null;
			try
			{
				var inputFileInfo = new FileInfo(excelFilePath);
				if (!inputFileInfo.Exists)
					throw new FileNotFoundException(string.Format("Input file \"{0}\" not found!", excelFilePath), excelFilePath);

				// Create a temporary copy of the input file, in case it's currently open in Excel/OpenOffice/LibreOffice.
				tempInputFile = inputFileInfo.CopyTo(Path.GetTempFileName(), true);
				using (var document = SpreadsheetDocument.Open(tempInputFile.FullName, false))
				{
					var allStats = GetTranslatedStats(GetSheet(document, statsSheetName));
					if (includeDifferencesOnly && !string.IsNullOrWhiteSpace(defaultsSheetName) &&
					    !StringEquals(defaultsSheetName, statsSheetName))
					{
						var allDefaults = GetTranslatedStats(GetSheet(document, defaultsSheetName));
						var statsKeys = allStats.Keys.ToArray();
						foreach (var key in statsKeys)
						{
							var statsEntry = allStats[key];
							Dictionary<string, string> defaultsEntry;
							if (allDefaults.TryGetValue(key, out defaultsEntry))
							{
								statsEntry = statsEntry
									.Except(defaultsEntry)
									.ToDictionary(x => x.Key, x => x.Value, StringComparer.InvariantCultureIgnoreCase);
								statsEntry[ColumnName] = key;

								if (statsEntry.Count == 0)
									allStats.Remove(key);
								else
									allStats[key] = statsEntry;
							}
						}
					}

					return allStats;
				}
			}
			finally
			{
				if (tempInputFile != null)
				{
					tempInputFile.Delete();
				}
			}
		}

		/// <summary>
		/// Retrieves a sheet from an Excel file, with each row keyed on the name column.
		/// </summary>
		/// <param name="document">The Excel document to read from.</param>
		/// <param name="statsSheetName">The name of the sheet to read from in the Excel file.</param>
		/// <returns></returns>
		private static Dictionary<string, Dictionary<string, string>> GetSheet(SpreadsheetDocument document,
			string statsSheetName)
		{
			var results = new Dictionary<string, Dictionary<string, string>>(StringComparer.InvariantCultureIgnoreCase);
			var wbPart = document.WorkbookPart;
			var sheet = wbPart.Workbook
				.Descendants<Sheet>()
				.FirstOrDefault(s => StringEquals(s.Name, statsSheetName));

			if (sheet == null)
				throw new NullReferenceException("Excel document does not contain a sheet named " + statsSheetName);

			var wsPart = (WorksheetPart) (wbPart.GetPartById(sheet.Id));

			var rawRows = wsPart
				.Worksheet
				.Descendants<Row>()
				.OrderBy(r => r.RowIndex.Value)
				.ToArray();

			var rows = rawRows
				.Select(row =>
					new
					{
						RowIndex = row.RowIndex.Value,
						Columns = row
							.Elements<Cell>()
							.Select(cell => new
							{
								Cell = cell,
								Column = ExtractColumnIndex(cell.CellReference)
							})
							.ToDictionary(x => x.Column, x => x.Cell, StringComparer.InvariantCultureIgnoreCase)
					})
				.OrderBy(x => x.RowIndex)
				.ToArray();

			if (!rows.Any())
				throw new Exception("Excel document contains no rows!");

			var headerAddresses = new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase);
			foreach (var col in rows.First().Columns)
			{
				var columnName = GetValue(wbPart, col.Value);
				if (string.IsNullOrWhiteSpace(columnName))
					continue;

				try
				{
					headerAddresses.Add(columnName, col.Key);
				}
				catch (ArgumentException ex)
				{
					throw new Exception(string.Format("Header column {0} has non-unique name: {1}", col.Key, columnName), ex);
				}
			}

			Func<Dictionary<string, Cell>, string, string> getValueWithRowAndColumnIndex = (r, ci) =>
			{
				Cell cell;
				return r.TryGetValue(ci, out cell)
					? GetValue(wbPart, cell)
					: null;
			};

			foreach (var r in rows.Skip(1))
			{
				var wholeRow = headerAddresses
					.ToDictionary(hc => hc.Key, hc => getValueWithRowAndColumnIndex(r.Columns, hc.Value),
						StringComparer.InvariantCultureIgnoreCase);

				if (!string.IsNullOrWhiteSpace(wholeRow[ColumnName]))
					results.Add(wholeRow[ColumnName], wholeRow);
			}

			return results;
		}

		/// <summary>
		/// Translates stats read from an Excel sheet into the game's internal representation.
		/// </summary>
		/// <param name="allStats">All stats read from an Excel sheet.</param>
		/// <returns>The translated stats from the Excel sheet.</returns>
		private static Dictionary<string, Dictionary<string, string>> GetTranslatedStats(
			Dictionary<string, Dictionary<string, string>> allStats)
		{
			var results = new Dictionary<string, Dictionary<string, string>>(StringComparer.InvariantCultureIgnoreCase);

			foreach (var plantStats in allStats)
			{
				var translatedStats = new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase);
				var isHydroPowerPlant = StringEquals(plantStats.Key, BuildingNameHydro);

				if (isHydroPowerPlant)
				{
					int damConstructionCost;
					if (!int.TryParse(plantStats.Value[ColumnCostPerCell], out damConstructionCost))
						throw new Exception(string.Format("Couldn't parse {0} column (integer), needed for calculating cost of {1}.",
							ColumnCostPerCell, BuildingNameHydro));

					damConstructionCost = (int) (damConstructionCost * DamConstructionCostMultiplier);

					results.Add(NetNameDam, new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase)
					{
						{ ColumnName, NetNameDam },
						{ ColumnConstructionCost, damConstructionCost.ToString(CultureInfo.InvariantCulture) },
					});
				}

				foreach (var statName in PlantStatNames)
				{
					var isInteger = true;
					var isValueResolved = false;
					var factor = 1.0;
					var key = statName;
					string value = null;

					if (StringEquals(statName, ColumnResourceType))
					{
						isValueResolved = true;
						isInteger = false;

						if (!plantStats.Value.TryGetValue(key, out value) || string.IsNullOrWhiteSpace(value))
							value = "None";
						value = string.Format("TransferManager.TransferReason.{0}", value);
					}
					else if (StringEquals(statName, ColumnElectricityProduction))
					{
						factor = ElectricityProductionMultiplier;
					}
					else if (StringEquals(statName, ColumnConstructionCost))
					{
						// Set constructionCost = 0 for the Hydro Power Plant itself.
						// The dam network's cost/cell is used instead.
						factor = isHydroPowerPlant ? 0 : 1;
					}
					else if (StringEquals(statName, ColumnMaintenanceCost))
					{
						factor = MaintenanceCostMultiplier;
					}
					else if (StringEquals(statName, ColumnWorkPlaceCount0))
					{
						key = ColumnUneducatedWorkers;
					}
					else if (StringEquals(statName, ColumnWorkPlaceCount1))
					{
						key = ColumnEducatedWorkers;
					}
					else if (StringEquals(statName, ColumnWorkPlaceCount2))
					{
						key = ColumnWellEducatedWorkers;
					}
					else if (StringEquals(statName, ColumnWorkPlaceCount3))
					{
						key = ColumnHighlyEducatedWorkers;
					}
					else if (StringEquals(statName, ColumnResourceCapacity))
					{
						isValueResolved = true;
						string consumptionString, capacityInWeeksString;
						int consumption, capacityInWeeks, totalCapacity;
						if (
							plantStats.Value.TryGetValue(ColumnResourceConsumption, out consumptionString) &&
							plantStats.Value.TryGetValue(ColumnResourceCapacityInWeeks, out capacityInWeeksString) &&
							TryParseInt(consumptionString, out consumption) &&
							TryParseInt(capacityInWeeksString, out capacityInWeeks))
						{
							totalCapacity = GetResourceTotalCapacity(capacityInWeeks, consumption);
						}
						else
						{
							totalCapacity = 0;
						}

						value = totalCapacity.ToString(CultureInfo.InvariantCulture);
					}

					if (!isValueResolved)
					{
						if (!plantStats.Value.TryGetValue(key, out value) || string.IsNullOrWhiteSpace(value))
							continue;

						// ReSharper disable once ConditionIsAlwaysTrueOrFalse
						if (isInteger)
						{
							int integerValue;
							if (!TryParseInt(value, out integerValue))
								throw new InvalidCastException(string.Format("Column {0} could not be read as a number.", key));

							integerValue = (int) Math.Round(factor * integerValue);
							value = integerValue.ToString(CultureInfo.InvariantCulture);
						}
					}

					translatedStats.Add(statName, value);
				}

				if (translatedStats.Count == 0)
					continue;

				results.Add(plantStats.Key, translatedStats);
			}

			return results;
		}

		/// <summary>
		/// Retrieves the code for the methods in the Rebuilder file.
		/// </summary>
		/// <param name="translatedStats">A dictionary of translatedStats keyed on power plant names.</param>
		/// <returns>A dictionary of code segments generated from the contents of the Excel file.</returns>
		private static Dictionary<string, string> GenerateRebuilderCode(
			Dictionary<string, Dictionary<string, string>> translatedStats)
		{
			var results = new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase);

			var sb = new StringBuilder();
			foreach (var plantStats in translatedStats)
			{
				if (StringEquals(NetNameDam, plantStats.Key))
				{
					results[TemplateModifyDam] = FormatString(DamFormatString, plantStats.Value).Trim();
				}
				else
				{
					var sbOnePlant = new StringBuilder();
					foreach (var statName in PlantStatNames)
					{
						string value;
						if (!plantStats.Value.TryGetValue(statName, out value))
							continue;
						sbOnePlant.AppendFormat("\t\t\t\tplant.m_{0} = {1};\n", statName, value);
					}

					var formatted = FormatString(PlantFormatString,
						new Dictionary<string, string>(StringComparer.InvariantCultureIgnoreCase)
						{
							{ ColumnName, plantStats.Key },
							{ TemplateSetStats, sbOnePlant.ToString().Trim() },
						});
					sb.Append(formatted);
				}
			}

			results[TemplateModifyPowerPlants] = sb.ToString().Trim();

			return results;
		}

		#region Excel File
		/// <summary>
		/// The name of the DEFAULTS sheet on the Excel file.
		/// </summary>
		private const string DefaultSheetName = "DEFAULTS";

		/// <summary>
		/// The construction cost (in dollars) of a single cell of a dam must
		/// be multiplied by this value to equal the program's internal cost.
		/// </summary>
		private const double DamConstructionCostMultiplier = 100.0;

		/// <summary>
		/// The maintenance cost (in dollars) of a power plant must
		/// be multiplied by this value to equal the program's internal cost.
		/// </summary>
		public const double MaintenanceCostMultiplier = 6.25;

		/// <summary>
		/// The electricity production (in megawatts) of a power plant must
		/// be multiplied by this value to equal to the program's internal production.
		/// </summary>
		public const double ElectricityProductionMultiplier = 62.5;

		/// <summary>
		/// The name of the hydroelectric power plant.
		/// </summary>
		private const string BuildingNameHydro = "Hydro Power Plant";

		/// <summary>
		/// The name of the dam net building.
		/// </summary>
		private const string NetNameDam = "Dam";

		#region Excel Columns
		private const string ColumnName = "name";
		private const string ColumnCostPerCell = "$/cell";
		private const string ColumnUneducatedWorkers = "uneducatedWorkers";
		private const string ColumnEducatedWorkers = "educatedWorkers";
		private const string ColumnWellEducatedWorkers = "wellEducatedWorkers";
		private const string ColumnHighlyEducatedWorkers = "highlyEducatedWorkers";
		private const string ColumnResourceCapacityInWeeks = "resourceCapacity (weeks)";
		#endregion Excel Columns

		#endregion Excel File

		#region Code Columns
		private const string ColumnConstructionCost = "constructionCost";
		private const string ColumnElectricityConsumption = "electricityConsumption";
		private const string ColumnElectricityProduction = "electricityProduction";
		private const string ColumnFireHazard = "fireHazard";
		private const string ColumnFireTolerance = "fireTolerance";
		private const string ColumnGarbageAccumulation = "garbageAccumulation";
		private const string ColumnMaintenanceCost = "maintenanceCost";
		private const string ColumnNoiseAccumulation = "noiseAccumulation";
		private const string ColumnNoiseRadius = "noiseRadius";
		private const string ColumnPollutionAccumulation = "pollutionAccumulation";
		private const string ColumnPollutionRadius = "pollutionRadius";
		private const string ColumnResourceCapacity = "resourceCapacity";
		private const string ColumnResourceConsumption = "resourceConsumption";
		private const string ColumnResourceType = "resourceType";
		private const string ColumnSewageAccumulation = "sewageAccumulation";
		private const string ColumnWaterConsumption = "waterConsumption";
		private const string ColumnWorkPlaceCount0 = "workPlaceCount0";
		private const string ColumnWorkPlaceCount1 = "workPlaceCount1";
		private const string ColumnWorkPlaceCount2 = "workPlaceCount2";
		private const string ColumnWorkPlaceCount3 = "workPlaceCount3";

		private static readonly string[] PlantStatNames =
		{
			ColumnConstructionCost,
			ColumnMaintenanceCost,
			ColumnElectricityConsumption,
			ColumnElectricityProduction,
			ColumnPollutionAccumulation,
			ColumnPollutionRadius,
			ColumnNoiseAccumulation,
			ColumnNoiseRadius,
			ColumnWaterConsumption,
			ColumnSewageAccumulation,
			ColumnGarbageAccumulation,
			ColumnResourceType,
			ColumnResourceConsumption,
			ColumnResourceCapacity,
			ColumnFireHazard,
			ColumnFireTolerance,
			ColumnWorkPlaceCount0,
			ColumnWorkPlaceCount1,
			ColumnWorkPlaceCount2,
			ColumnWorkPlaceCount3,
		};
		#endregion Code Columns

		#region Template
		/// <summary>
		/// The namespace token in the templates.
		/// </summary>
		private const string TemplateAddonNamespace = "namespace";

		/// <summary>
		/// The addonName token in the templates.
		/// </summary>
		private const string TemplateAddonName = "addonName";

		/// <summary>
		/// The description token in the templates.
		/// </summary>
		private const string TemplateDescription = "description";

		/// <summary>
		/// The setStats token in the PlantFormatString template.
		/// </summary>
		private const string TemplateSetStats = "setStats";

		/// <summary>
		/// The modifyPowerPlants token in the Rebuilder.cs template.
		/// </summary>
		private const string TemplateModifyPowerPlants = "modifyPowerPlants";

		/// <summary>
		/// The modifyDam token in the Rebuilder.cs template.
		/// </summary>
		private const string TemplateModifyDam = "modifyDam";

		/// <summary>
		/// The format string for a power plant modification section.
		/// </summary>
		private const string PlantFormatString = @"
			#region ${name}
			if (string.Equals(name, ""${name}"", StringComparison.InvariantCultureIgnoreCase) ||
				string.Equals(localizedName, ""${name}"", StringComparison.InvariantCultureIgnoreCase))
			{
				${setStats}
				return;
			}
			#endregion ${name}
";

		/// <summary>
		/// The format string for a dam modification section.
		/// </summary>
		private const string DamFormatString = @"
			#region ${name}
			dam.m_constructionCost = ${constructionCost};
			#endregion ${name}
";
		#endregion Template

		#region Helpers
		/// <summary>
		/// Escapes the provided <paramref name="text"/> so that it can be surrounded by quotes in code.
		/// </summary>
		/// <param name="text">A text string.</param>
		/// <returns>The escaped string.</returns>
		private static string EscapeVerbatimString(string text)
		{
			return text.Replace("\"", "\"\"");
		}

		/// <summary>
		/// Performs a keyword-based string format rather than using numeric indexes.
		/// </summary>
		/// <example>
		/// Replace ${a} and ${b} with <paramref name="parameters"/>["a"] and <paramref name="parameters"/>["b].
		/// </example>
		/// <param name="formatString">A format string in which the tokens to be replaced are in the form "${key}".</param>
		/// <param name="parameters">A dictionary of keys with values to be inserted into the <paramref name="formatString"/>.</param>
		/// <returns></returns>
		private static string FormatString(string formatString, IDictionary<string, string> parameters)
		{
			var formatted = Regex.Replace(formatString, @"\$\{(\w+)\}", m => parameters[m.Groups[1].Value]);
			return formatted;
		}

		/// <summary>
		/// Extracts the column index from a cell address.
		/// </summary>
		/// <param name="address">A cell address in the form A1, B2, C3, etc.</param>
		/// <returns>The column index (A, B, C, etc).</returns>
		private static string ExtractColumnIndex(string address)
		{
			var m = Regex.Match(address, @"\A([a-zA-Z]+)");
			return m.Success
				? m.Groups[1].Value
				: null;
		}

		/// <summary>
		/// Converts an alphabetic <paramref name="columnIndex"/> to a numeric index.
		/// </summary>
		/// <param name="columnIndex">An alphabetic column index (A, B, C, etc).</param>
		/// <returns>The associated numeric column index.</returns>
		public static int ExcelColumnNameToNumber(string columnIndex)
		{
			if (string.IsNullOrEmpty(columnIndex))
				throw new ArgumentNullException("columnIndex");

			columnIndex = columnIndex.ToUpperInvariant();

			var sum = 0;
			for (var i = 0; i < columnIndex.Length; i++)
			{
				sum *= 26;
				sum += (i - 'A' + 1);
			}

			return sum - 1;
		}

		/// <summary>
		/// Returns the trimmed <paramref name="value"/>, or <see cref="string.Empty"/> if null.
		/// </summary>
		/// <param name="value">A string value to be trimmed</param>
		/// <returns>The trimmed <paramref name="value"/>, or <see cref="string.Empty"/> if null</returns>
		private static string TrimString(string value)
		{
			return string.IsNullOrWhiteSpace(value) ? string.Empty : value.Trim();
		}

		/// <summary>
		/// Returns true if <paramref name="s1"/> and <paramref name="s2"/> are equal, ignoring case and culture.
		/// </summary>
		/// <param name="s1">The first string.</param>
		/// <param name="s2">The second string.</param>
		/// <returns>True if the two strings are equal, ignoring case and culture.</returns>
		private static bool StringEquals(string s1, string s2)
		{
			return string.Equals(s1, s2, StringComparison.InvariantCultureIgnoreCase);
		}

		/// <summary>
		/// Parses the provided <paramref name="text"/> as a <see cref="double"/>,
		/// then rounds it to an <see cref="int"/>, returning true on success.
		/// </summary>
		/// <param name="text">A text string, possibly containing a number.</param>
		/// <param name="value">The output integer.</param>
		/// <returns>True if the number could be parsed and rounded.</returns>
		private static bool TryParseInt(string text, out int value)
		{
			double dValue;
			if (double.TryParse(text, NumberStyles.Number, CultureInfo.InvariantCulture, out dValue))
			{
				value = (int) Math.Round(dValue);
				return true;
			}

			value = default(int);
			return false;
		}

		/// <summary>
		/// Calculates the total resource capacity from the <paramref name="capacityInWeeks"/> and <paramref name="consumption"/> rate.
		/// </summary>
		/// <param name="capacityInWeeks"></param>
		/// <param name="consumption"></param>
		/// <returns></returns>
		private static int GetResourceTotalCapacity(int capacityInWeeks, int consumption)
		{
			return Math.Max(((capacityInWeeks * 16) - 8) * consumption, 0);
		}

		/// <summary>
		/// Gets the value in the cell, or the shared string referenced by the cell.
		/// </summary>
		/// <param name="workbookPart">The workbook part in which the <see cref="SharedStringTablePart"/> can be found</param>
		/// <param name="cell">The cell to extract from.</param>
		/// <returns>The contents of the cell, or <see cref="string.Empty"/> if null.</returns>
		public static string GetValue(WorkbookPart workbookPart, Cell cell)
		{
			if (cell.DataType == null || !cell.DataType.HasValue)
				return TrimString(cell.InnerText);

			switch (cell.DataType.Value)
			{
				case CellValues.Boolean:
					return (cell.InnerText != "0").ToString(CultureInfo.InvariantCulture);
				case CellValues.SharedString:
				{
					var stringTable = workbookPart.GetPartsOfType<SharedStringTablePart>().FirstOrDefault();
					return stringTable == null
						? null
						: TrimString(stringTable.SharedStringTable.ElementAt(Int32.Parse(cell.CellValue.Text)).InnerText);
				}
				default:
				{
					return TrimString(cell.CellValue.Text);
				}
			}
		}
		#endregion Helpers
	}
}